from __future__ import print_function
import socket
import multiprocessing
import atexit
from . import olib


def close_socket(sock):
    sock.shutdown(socket.SHUT_RDWR)
    sock.close()


def resolve(rerouter, conn, resolve_callback=lambda q, a: (q, a)):
    try:
        while True:
            addr = conn.recv(1024).strip()
            if not addr:
                # connection ended
                return
            if addr == 'get *':
                conn.sendall("500 Request key is not an email address\n")
            else:
                result = rerouter.run(addr)
                resolve_callback(addr, result)
                conn.sendall("{0}\n".format(result))
    except socket.timeout:
        return
    except BaseException as err:
        # todo log
        conn.sendall("500 {0}".format(err))


def daemonize_server(rerouter, host, port, resolver=resolve):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind((host, port))
    sock.listen(1)
    atexit.register(close_socket, sock=sock)
    while True:
        conn, address = sock.accept()
        process = multiprocessing.Process(target=resolver,
                                          args=(rerouter, conn))
        process.daemon = True
        process.start()


def client(host, port):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.connect((host, port))
    atexit.register(close_socket, sock=sock)
    while True:
        addr = olib.cross_input("Enter an email address: ")
        if addr == 'get *':
            print("500 Request key is not an email address")
        else:
            sock.sendall(addr)
            print(sock.recv(1024).strip())



import os
try:
    import ConfigParser as configparser
except ImportError:
    import configparser
from yaml import load
from .olib import find_file, find_files_with_suffix
from .custom_exceptions import ConfigIntegrityError, ConfigNotFoundError


class ConfigIntegrityChecker(object):
    def __init__(self, ref_config, other_config):
        self._ref_reader = config_reader(ref_config)
        self._oth_reader = config_reader(other_config)
        self.ref_config = ref_config
        self.other_config = other_config

    def _verify_sections(self):
        if set(self._ref_reader.sections()).difference(
                set(self._oth_reader.sections())):
            raise RuntimeError

    def _verify_options(self):
        for section in self._ref_reader.sections():
            if set(self._ref_reader.options(section)).difference(
                    set(self._oth_reader.options(section))):
                raise RuntimeError

    def verify(self):
        try:
            self._verify_sections()
            self._verify_options()
        except RuntimeError:
            raise ConfigIntegrityError(
                "{loc} does not match the reference config file {ref}\n"
                "Is your local config up to date?".format(
                    loc=self.other_config, ref=self.ref_config))


def config_reader(conf_path):
    """
    Reads a conf file and returns the onfigParser object
    """
    config = configparser.ConfigParser()
    config.read(conf_path)
    return config


def get_conffile(conf_path, prefix='', suffix=".ini"):
    """
    Finds conf file in `conf_path`
    If `conf_path` is file, returns `conf_path`
    Else it searches to find a file with `prefix``suffix` format
    """
    if os.path.isfile(conf_path):
        return conf_path

    found_file = find_file(conf_path, "{0}{1}".format(prefix, suffix))
    if not found_file:
        raise ConfigNotFoundError('No configuration file found in {0}'
                                  .format(conf_path))
    return found_file


def yaml_loader(yaml_path):
    """
    Yields yaml files
    Supports folder lookup
    :param yaml_path: path pointing to a yaml file or a dir with yaml files
    :return: yields open file
    """
    if not os.path.exists(yaml_path):
        return

    if os.path.isdir(yaml_path):
        all_files = sorted(find_files_with_suffix(yaml_path, "yml"))
    else:
        all_files = [yaml_path]

    for filename in all_files:
        with open(filename, 'r') as yaml_file:
            yield yaml_file


def load_yamls(yaml_path):
    """
    Returns a dict with mappings defined in yaml file(s)
    :param yaml_path: path pointing to a yaml file or a dir with yaml files
    :rtype: dict
    """
    yaml_mapping = dict()
    for f in yaml_loader(yaml_path):
        yaml_mapping.update(load(f))
    return yaml_mapping



from dns.resolver import Resolver
import re


class OnionServiceLookup(object):
    def __init__(self, config):
        self.config = config
        self.resolver = Resolver()
        self.resolver.nameservers = self.config.get("RESOLVER",
                                                    "resolver_ip").split(",")
        self.resolver.port = int(self.config.get("RESOLVER", "resolver_port"))

    @property
    def srv_record(self):
        return self.config.get("DNS", "srv_record")

    @property
    def use_tcp(self):
        return self.config.get("RESOLVER", "tcp") == "True"

    @staticmethod
    def is_onion(response):
        return re.search(r"onion\.$", response) is not None

    def _craft_format_record(self, domain):
        return "{0}{1}".format(self.srv_record, domain)

    def _map_answers(self, answers):
        return tuple(filter(self.is_onion, (str(x.target) for x in answers)))

    @staticmethod
    def _craft_output(answers):
        if answers:
            return tuple("{data}".format(data=str(x).rstrip('.'))
                         for x in answers)
        else:
            return tuple("")

    def lookup(self, domain):
        query = self.resolver.query(self._craft_format_record(domain),
                                    "SRV", tcp=self.use_tcp)
        return self._craft_output(self._map_answers(query))
